#include "tsi_object.h"
#include "tsi_processing.h"
#include "tsi_driver.h"
#include "tsi_plugin.h"

/* Configurations -----------------------------------------------------------*/
/** Plugin call priority(0-7). Lower value means higher priority. */
#define TSI_PLUGIN_PRIORITY                 "0"

/** Plugin version string. */
#define TSI_PLUGIN_VERSION                  "v1.0"

/* USER CONFIGURATION BEGIN */
/** Use test functions. */
#define TSI_COMMNOISE_TEST                  (0U)
/* USER CONFIGURATION END */
/* Defines ------------------------------------------------------------------*/
#define TSI_IS_SELF_POS_COMMNOISE           (0x01U)
#define TSI_IS_MUTUAL_POS_COMMNOISE         (0x10U)
#define TSI_IS_SELF_NEG_COMMNOISE           (0x01U)
#define TSI_IS_MUTUAL_NEG_COMMNOISE         (0x10U)

/* Function prototypes ------------------------------------------------------*/
static void TSI_Widget_ValueUpdateCallback(TSI_LibHandleTypeDef *handle);
static void TSI_Widget_BaselineInit(TSI_LibHandleTypeDef *handle);
static uint8_t TSI_Widget_isPosComNoise(TSI_LibHandleTypeDef *handle);
static uint8_t TSI_Widget_isNegComNoise(TSI_LibHandleTypeDef *handle);
#if TSI_COMMNOISE_TEST
    static void TSI_Widget_PosTest(TSI_LibHandleTypeDef *handle);
    static void TSI_Widget_NegTest(TSI_LibHandleTypeDef *handle);
#endif

/* Variables ----------------------------------------------------------------*/
static uint8_t negCommNoiseTimeCnt = 0U;

/* Function implementations -------------------------------------------------*/
static void TSI_Widget_ValueUpdateCallback(TSI_LibHandleTypeDef *handle)
{
    TSI_ScanGroupTypeDef *group;
    group = handle->driver->scanGroups;

    /* Postive ComNoise Detect */
    TSI_FOREACH_OBJ(uint16_t *, snsId, group->sensors, group->size) {
        TSI_SensorTypeDef *sensor = TSI_SensorPointers[*snsId];
        TSI_WidgetTypeDef *widget = sensor->meta->parent;
        if(TSI_WIDGET_IS_SELF_CAP(widget)) {
            int diff = ((int32_t)sensor->rawCount[0u] - (int32_t)sensor->baseline[0u]);
            if(diff > ((int32_t)(widget->detConf.activeTh + widget->detConf.activeHys))) {
                if((TSI_Widget_isPosComNoise(handle) & TSI_IS_SELF_POS_COMMNOISE) == 1U) {
                    TSI_Widget_BaselineInit(handle);
                }
            }
            else {
                /* Do nothing. */
            }
        }
        else if(TSI_WIDGET_IS_MUTUAL_CAP(widget)) {
            /** @todo Mutual Cap implementation. */
        }
        else {
            /* Do nothing. */
        }
    }
    TSI_FOREACH_END()

    /* Negative ComNoise Detect */
    if((TSI_Widget_isNegComNoise(handle) & TSI_IS_SELF_NEG_COMMNOISE) == 1u) {
        if(++negCommNoiseTimeCnt > 5U) {
            TSI_Widget_BaselineInit(handle);
            negCommNoiseTimeCnt = 0U;
        }
    }
    else {
        negCommNoiseTimeCnt = 0U;
    }
}

static void TSI_Widget_BaselineInit(TSI_LibHandleTypeDef *handle)
{
    TSI_ScanGroupTypeDef *group;
    group = handle->driver->scanGroups;

    TSI_FOREACH_OBJ(uint16_t *, snsId, group->sensors, group->size) {
        TSI_SensorTypeDef *sensor = TSI_SensorPointers[*snsId];
        TSI_WidgetTypeDef *widget = sensor->meta->parent;
        if(TSI_WIDGET_IS_SELF_CAP(widget)) {
            TSI_Baseline_Init(sensor, &widget->detConf);
            sensor->diffCount = 0U;
#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))
            sensor->ltaDiffCount = 0U;
#endif
        }
        else if(TSI_WIDGET_IS_MUTUAL_CAP(widget)) {
            /** @todo Mutual Cap implementation. */
        }
        else {
            /* Do nothing. */
        }
    }
    TSI_FOREACH_END();
}

static uint8_t TSI_Widget_isPosComNoise(TSI_LibHandleTypeDef *handle)
{
    uint8_t isPosComNoise = (TSI_IS_SELF_POS_COMMNOISE | TSI_IS_MUTUAL_POS_COMMNOISE);

    TSI_ScanGroupTypeDef *group;
    group = handle->driver->scanGroups;

    TSI_FOREACH_OBJ(uint16_t *, snsId, group->sensors, group->size) {
        const TSI_SensorTypeDef *sensor = TSI_SensorPointers[*snsId];
        TSI_WidgetTypeDef *widget = sensor->meta->parent;
        int diff = ((int32_t)sensor->rawCount[0u] - (int32_t)sensor->baseline[0u]);
        if(TSI_WIDGET_IS_SELF_CAP(widget)) {
            if(diff < ((int32_t)(widget->detConf.noiseTh))) {
                isPosComNoise &= ~(TSI_IS_SELF_POS_COMMNOISE);
            }
        }
        else if(TSI_WIDGET_IS_MUTUAL_CAP(widget)) {
            if(diff < ((int32_t)(widget->detConf.noiseTh))) {
                isPosComNoise &= ~(TSI_IS_MUTUAL_POS_COMMNOISE);
            }
        }
        else {
            /* Do nothing. */
        }
    }
    TSI_FOREACH_END();

    return isPosComNoise;
}

static uint8_t TSI_Widget_isNegComNoise(TSI_LibHandleTypeDef *handle)
{
    uint8_t isNegComNoise = (TSI_IS_SELF_NEG_COMMNOISE | TSI_IS_MUTUAL_NEG_COMMNOISE);

    TSI_ScanGroupTypeDef *group;
    group = handle->driver->scanGroups;

    TSI_FOREACH_OBJ(uint16_t *, snsId, group->sensors, group->size) {
        const TSI_SensorTypeDef *sensor = TSI_SensorPointers[*snsId];
        TSI_WidgetTypeDef *widget = sensor->meta->parent;
        int diff = (int32_t)sensor->rawCount[0u] - (int32_t)sensor->baseline[0u];
        if(TSI_WIDGET_IS_SELF_CAP(widget)) {
            if(diff > -((int32_t)widget->detConf.negNoiseTh)) {
                isNegComNoise &= ~(TSI_IS_SELF_NEG_COMMNOISE);
            }
        }
        else if(TSI_WIDGET_IS_MUTUAL_CAP(widget)) {
            if(diff > -((int32_t)widget->detConf.negNoiseTh)) {
                isNegComNoise &= ~(TSI_IS_MUTUAL_NEG_COMMNOISE);
            }
        }
        else {
            /* Do nothing. */
        }
    }
    TSI_FOREACH_END();

    return isNegComNoise;
}

#if TSI_COMNOISE_TEST
static void TSI_Widget_PosTest(TSI_LibHandleTypeDef *handle)
{
    TSI_ScanGroupTypeDef *group;
    group = handle->driver->scanGroups;

    TSI_FOREACH_OBJ(uint16_t *, snsId, group->sensors, group->size) {
        TSI_SensorTypeDef *sensor = TSI_SensorPointers[*snsId];
        TSI_WidgetTypeDef *widget = sensor->meta->parent;

        if(TSI_WIDGET_IS_SELF_CAP(widget)) {
            sensor->bslnVar.sensorBuffer[0] = sensor->bslnVar.sensorBuffer[0] + 200;
        }
        else if(TSI_WIDGET_IS_MUTUAL_CAP(widget)) {
            /** @todo Mutual Cap implementation. */
        }
        else {
            /* Do nothing. */
        }
    }
    TSI_FOREACH_END();
}

static void TSI_Widget_NegTest(TSI_LibHandleTypeDef *handle)
{
    TSI_ScanGroupTypeDef *group;
    group = handle->driver->scanGroups;

    TSI_FOREACH_OBJ(uint16_t *, snsId, group->sensors, group->size) {
        TSI_SensorTypeDef *sensor = TSI_SensorPointers[*snsId];
        TSI_WidgetTypeDef *widget = sensor->meta->parent;

        if(TSI_WIDGET_IS_SELF_CAP(widget)) {
            sensor->bslnVar.sensorBuffer[0] = sensor->bslnVar.sensorBuffer[0] - 200;
        }
        else if(TSI_WIDGET_IS_MUTUAL_CAP(widget)) {
            /** @todo Mutual Cap implementation. */
        }
        else {
            /* Do nothing. */
        }
    }
    TSI_FOREACH_END();
}
#endif

/* Plugin registration ------------------------------------------------------*/
TSI_PLUGIN(TSI_PLUGIN_PRIORITY)
{
    NULL,                               /* initCompleted */
    NULL,                               /* deInitCompleted */
    NULL,                               /* started */
    NULL,                               /* stopped */
    NULL,                               /* widgetInitCompleted */
    NULL,                               /* widgetScanCompleted */
    TSI_Widget_ValueUpdateCallback,     /* widgetValueUpdated */
    NULL,                               /* widgetStatusUpdated */
    NULL,                               /* getInitScanBufferAndCount */
    NULL,                               /* processInitScanValue */
};
