#include "tsi_processing.h"
#include "tsi_filter.h"
#include "tsi.h"

/* Private defines ----------------------------------------------------------*/
#define TSI_SENSOR_STATUS_ACTIVE        (uint8_t)(0x1U)
#define TSI_SENSOR_STATUS_PROX          (uint8_t)(0x2U)

#define TSI_BASELINE_MODE_NORMAL        (uint8_t)(0x0U)
#define TSI_BASELINE_MODE_LTA           (uint8_t)(0x1U)

/* Private function prototypes ----------------------------------------------*/
/* Widget */
TSI_STATIC void TSI_Widget_ProcessDiffAndBaseline(TSI_LibHandleTypeDef *handle, TSI_WidgetTypeDef *widget);
TSI_STATIC void TSI_Widget_ProcessStatusAndBaseline(TSI_LibHandleTypeDef *handle, TSI_WidgetTypeDef *widget);
TSI_STATIC void TSI_Widget_ProcessPrivateData(TSI_LibHandleTypeDef *handle, TSI_WidgetTypeDef *widget);
/* Sensor */
TSI_STATIC void TSI_Sensor_ProcessDiffAndBaseline(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf);
TSI_STATIC void TSI_Sensor_ProcessStatusAndBaseline(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf);
/* Status */
TSI_STATIC void TSI_Sensor_UpdateStatus(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf,
                                        uint8_t type);
/* Normal baseline */
TSI_STATIC void TSI_NormalBaseline_Init(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf);
TSI_STATIC void TSI_NormalBaseline_Update(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf);
#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))
    /* LTA baseline */
    TSI_STATIC void TSI_LTABaseline_Init(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf);
    TSI_STATIC void TSI_LTABaseline_Update(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf);
    TSI_STATIC void TSI_Baseline_ModeJudge(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf);
#endif  /* if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U) */
/* DiffCount */
TSI_STATIC void TSI_Sensor_UpdateDiffCount(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf);

/* Widget API implementations -----------------------------------------------*/
TSI_RetCode TSI_Widget_InitAll(TSI_LibHandleTypeDef *handle)
{
    TSI_RetCode res;

    TSI_FOREACH_OBJ(TSI_WidgetTypeDef **, ppWidget, handle->widgets,
                    handle->widgetNum) {
        res = TSI_Widget_Init(handle, *ppWidget);
        if(res != TSI_PASS) { return res; }
    }
    TSI_FOREACH_END()

    return TSI_PASS;
}

TSI_RetCode TSI_Widget_Init(TSI_LibHandleTypeDef *handle, TSI_WidgetTypeDef *widget)
{
    TSI_DetectConfTypeDef *widgetDetConf = &widget->detConf;
    uint16_t realSnsNum;

    TSI_UNUSED(handle)

    /* If it is self-cap parallel widget, we shall only init the first sensor. */
    if(TSI_WIDGET_IS_SELF_CAP(widget) && widget->meta->dedicatedScanGroup != NULL) {
        realSnsNum = 1U;
    }
    else {
        realSnsNum = widget->meta->sensorNum;
    }

    TSI_FOREACH_OBJ(TSI_SensorTypeDef *, pSensor, widget->meta->sensors,
                    realSnsNum) {
        TSI_DetectConfTypeDef *detConf = pSensor->meta->detConf;
        if(detConf == NULL) { detConf = widgetDetConf; }
        TSI_ASSERT(detConf);
        TSI_Sensor_Init(pSensor, detConf);
    }
    TSI_FOREACH_END()

    /* Call user callback for customized algorithms. */
    if(handle->cb.widgetInitCompleted != NULL) {
        handle->cb.widgetInitCompleted(handle, widget);
    }

    return TSI_PASS;
}

TSI_RetCode TSI_Widget_EnableAll(TSI_LibHandleTypeDef *handle)
{
    TSI_RetCode res;
    TSI_FOREACH_OBJ(TSI_WidgetTypeDef **, ppWidget, handle->widgets,
                    handle->widgetNum) {
        res = TSI_Widget_Enable(handle, *ppWidget);
        if(res != TSI_PASS) {
            return res;
        }
    }
    TSI_FOREACH_END()
    return TSI_PASS;
}

TSI_RetCode TSI_Widget_Enable(TSI_LibHandleTypeDef *handle, TSI_WidgetTypeDef *widget)
{
    return TSI_Drv_EnableWidget(handle->driver, widget);
}

TSI_RetCode TSI_Widget_DisableAll(TSI_LibHandleTypeDef *handle)
{
    TSI_RetCode res;
    TSI_FOREACH_OBJ(TSI_WidgetTypeDef **, ppWidget, handle->widgets,
                    handle->widgetNum) {
        res = TSI_Widget_Disable(handle, *ppWidget);
        if(res != TSI_PASS) {
            return res;
        }
    }
    TSI_FOREACH_END()
    return TSI_PASS;
}

TSI_RetCode TSI_Widget_Disable(TSI_LibHandleTypeDef *handle, TSI_WidgetTypeDef *widget)
{
    return TSI_Drv_DisableWidget(handle->driver, widget);
}

void TSI_Widget_UpdateAll(TSI_LibHandleTypeDef *handle)
{
    /* Call user callback for customized algorithms. */
    if(handle->cb.widgetScanCompleted != NULL) {
        handle->cb.widgetScanCompleted(handle);
    }

    TSI_FOREACH_OBJ(TSI_WidgetTypeDef **, ppWidget, handle->widgets,
                    handle->widgetNum) {
        if((*ppWidget)->enable == TSI_WIDGET_ENABLE) {
            /* Update diffcount and baseline */
            TSI_Widget_ProcessDiffAndBaseline(handle, *ppWidget);
        }
    }
    TSI_FOREACH_END()

    /* Call user callback for customized algorithms. */
    if(handle->cb.widgetValueUpdated != NULL) {
        handle->cb.widgetValueUpdated(handle);
    }

    TSI_FOREACH_OBJ(TSI_WidgetTypeDef **, ppWidget, handle->widgets,
                    handle->widgetNum) {
        if((*ppWidget)->enable == TSI_WIDGET_ENABLE) {
            /* Update sensor status and baseline mode */
            TSI_Widget_ProcessStatusAndBaseline(handle, *ppWidget);
            /* Update widget-specified data */
            TSI_Widget_ProcessPrivateData(handle, *ppWidget);
        }
    }
    TSI_FOREACH_END()

    /* Call user callback for customized algorithms. */
    if(handle->cb.widgetStatusUpdated != NULL) {
        handle->cb.widgetStatusUpdated(handle);
    }
}

void TSI_Widget_UpdateSelfCapButton(TSI_SelfCapButtonTypeDef *button)
{
    button->buttonStat = button->base.base.status;
}

void TSI_Widget_UpdateSelfCapProximity(TSI_SelfCapProximityTypeDef *proximity)
{
    if(proximity->base.base.status == 1U) {
        /* Proximity widget is activated when any sensor is activated */
        proximity->proximityStat = 2U;
    }
    else {
        TSI_MetaWidgetTypeDef *metaWidget = proximity->base.base.meta;
        TSI_FOREACH_OBJ(TSI_SensorTypeDef *, pSensor, metaWidget->sensors,
                        metaWidget->sensorNum) {
            if((pSensor->status & TSI_SENSOR_STATUS_PROX) != 0U) {
                /* Proximity widget is in proximity status when any sensor is
                   in proximity status */
                proximity->proximityStat = 1U;
                return;
            }
        }
        TSI_FOREACH_END()

        proximity->proximityStat = 0U;
    }
}

void TSI_Widget_UpdateSelfCapSlider(TSI_SelfCapSliderTypeDef *slider)
{
    TSI_WidgetTypeDef *sliderBase = (TSI_WidgetTypeDef *)slider;

    if(sliderBase->status == 1U) {
        uint32_t peakIndex = 0U;
        uint16_t peakSignalVal = 0U;
        uint16_t centroidCalcData[3];
        int32_t centroid;
        const TSI_MetaWidgetTypeDef *meta = sliderBase->meta;

        int32_t mul = (TSI_SLIDER_RESOLUTION / (meta->sensorNum - 1U)) << 8U;

        /* Slider widget is activated when any sensor is activated */
        slider->sliderStat = 1U;

        /*
            Calculate the slider postion using centroid algorithm:
            1. Find the sensor X with the highest signal, which is the center of the touched area;
            2. Calcuate centroid using X's neighbour sensor;
        */

        /* Find peak value and corresponding sensor */
        TSI_FOREACH_OBJ(TSI_SensorTypeDef *, pSensor, meta->sensors,
                        meta->sensorNum) {
            TSI_DetectConfTypeDef *detConf;
            uint16_t threshold;

            /* Get detect configuration */
            detConf = pSensor->meta->detConf;
            if(detConf == NULL) { detConf = &pSensor->meta->parent->detConf; }

            /* Calclate threshold */
            threshold = detConf->activeTh - detConf->activeHys;

            if(pSensor->diffCount > threshold &&
                    pSensor->diffCount > peakSignalVal) {
                /* Record peak index */
                /** @todo Peak jitter ? */
                peakIndex = idx + 1U;
                peakSignalVal = (uint16_t)pSensor->diffCount;
            }
        }
        TSI_FOREACH_END()

        if(peakIndex != 0U) {
            /* Get neighbour sensors signal value */
            TSI_SensorTypeDef *sensor = meta->sensors;
            if(peakIndex == 1U) {
                centroidCalcData[0] = 0U;
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = 0U;
                if(sensor[1].diffCount > 0) {
                    centroidCalcData[2] = (uint16_t)sensor[1].diffCount;
                }
            }
            else if(peakIndex == meta->sensorNum) {
                centroidCalcData[0] = 0U;
                if(sensor[peakIndex - 2U].diffCount > 0) {
                    centroidCalcData[0] = (uint16_t)sensor[peakIndex - 2U].diffCount;
                }
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = 0;
            }
            else {
                centroidCalcData[0] = 0U;
                if(sensor[peakIndex - 2U].diffCount > 0) {
                    centroidCalcData[0] = (uint16_t)sensor[peakIndex - 2U].diffCount;
                }
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = 0U;
                if(sensor[peakIndex].diffCount > 0) {
                    centroidCalcData[2] = (uint16_t)sensor[peakIndex].diffCount;
                }
            }

            /* Calculate centroid */
            centroid = ((int32_t)centroidCalcData[2]
                        - (int32_t)centroidCalcData[0]) * mul /
                       ((int32_t)centroidCalcData[0]
                        + (int32_t)centroidCalcData[1]
                        + (int32_t)centroidCalcData[2]);
            centroid += (peakIndex - 1U) * mul;
            centroid /= (int32_t)256;
            slider->centerPos = (centroid > 0) ? (uint8_t)centroid : 0U;
        }
    }
    else {
        slider->sliderStat = 0U;
    }
}

void TSI_Widget_UpdateSelfCapTouchPad(TSI_SelfCapTouchPadTypeDef *pad)
{
    /** @todo IMPLEMENTATION */
}

void TSI_Widget_UpdateMutualCapButton(TSI_MutualCapButtonTypeDef *button)
{
    button->buttonStat = button->base.base.status;
}

void TSI_Widget_UpdateMutualCapSlider(TSI_MutualCapSliderTypeDef *slider)
{
    TSI_WidgetTypeDef *sliderBase = (TSI_WidgetTypeDef *)slider;

    if(sliderBase->status == 1U) {
        uint32_t peakIndex = 0U;
        uint16_t peakSignalVal = 0U;
        uint16_t centroidCalcData[3];
        int32_t centroid;
        const TSI_MetaWidgetTypeDef *meta = sliderBase->meta;

        int32_t mul = (TSI_SLIDER_RESOLUTION / (meta->sensorNum - 1U)) << 8U;

        /* Slider widget is activated when any sensor is activated */
        slider->sliderStat = 1U;

        /*
            Calculate the slider postion using centroid algorithm:
            1. Find the sensor X with the highest signal, which is the center of the touched area;
            2. Calcuate centroid using X's neighbour sensor;
        */

        /* Find peak value and corresponding sensor */
        TSI_FOREACH_OBJ(TSI_SensorTypeDef *, pSensor, meta->sensors,
                        meta->sensorNum) {
            TSI_DetectConfTypeDef *detConf;
            uint16_t threshold;

            /* Get detect configuration */
            detConf = pSensor->meta->detConf;
            if(detConf == NULL) { detConf = &pSensor->meta->parent->detConf; }

            /* Calclate threshold */
            threshold = detConf->activeTh - detConf->activeHys;

            if(pSensor->diffCount > threshold &&
                    pSensor->diffCount > peakSignalVal) {
                /* Record peak index */
                /** @todo Peak jitter ? */
                peakIndex = idx + 1U;
                peakSignalVal = (uint16_t)pSensor->diffCount;
            }
        }
        TSI_FOREACH_END()

        if(peakIndex != 0U) {
            /* Get neighbour sensors signal value */
            TSI_SensorTypeDef *sensor = meta->sensors;
            if(peakIndex == 1U) {
                centroidCalcData[0] = 0U;
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = 0U;
                if(sensor[1].diffCount > 0) {
                    centroidCalcData[2] = (uint16_t)sensor[1].diffCount;
                }
            }
            else if(peakIndex == meta->sensorNum) {
                centroidCalcData[0] = 0U;
                if(sensor[peakIndex - 2U].diffCount > 0) {
                    centroidCalcData[0] = (uint16_t)sensor[peakIndex - 2U].diffCount;
                }
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = 0;
            }
            else {
                centroidCalcData[0] = 0U;
                if(sensor[peakIndex - 2U].diffCount > 0) {
                    centroidCalcData[0] = (uint16_t)sensor[peakIndex - 2U].diffCount;
                }
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = 0U;
                if(sensor[peakIndex].diffCount > 0) {
                    centroidCalcData[2] = (uint16_t)sensor[peakIndex].diffCount;
                }
            }

            /* Calculate centroid */
            centroid = ((int32_t)centroidCalcData[2]
                        - (int32_t)centroidCalcData[0]) * mul /
                       ((int32_t)centroidCalcData[0]
                        + (int32_t)centroidCalcData[1]
                        + (int32_t)centroidCalcData[2]);
            centroid += (peakIndex - 1U) * mul;
            centroid /= (int32_t)256;
            slider->centerPos = (centroid > 0) ? (uint8_t)centroid : 0U;
        }
    }
    else {
        slider->sliderStat = 0U;
    }
}

/* Sensor APIs declaration --------------------------------------------------*/
void TSI_Sensor_Init(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    /* Init sensor filter buffers */
    TSI_Filter_Init(sensor);

    /* Init sensor baseline */
    TSI_Baseline_Init(sensor, detConf);

    /* Reset sensor status and debounce counter */
    sensor->status = 0U;
    memset(sensor->meta->debArray, detConf->onDebounce, sensor->meta->debArraySize);
}

/* Baseline APIs declaration ------------------------------------------------*/
void TSI_Baseline_Init(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    TSI_NormalBaseline_Init(sensor, detConf);
#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))
    TSI_LTABaseline_Init(sensor, detConf);
    sensor->bslnVar.bslnMode = TSI_BASELINE_MODE_NORMAL;
#endif  /* if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U) */
}

void TSI_Baseline_Update(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))
    uint8_t bslnMode = sensor->bslnVar.bslnMode;
#endif  /* if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U) */

    /* Update baseline. */
#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))
    if(bslnMode == TSI_BASELINE_MODE_NORMAL) {
        TSI_NormalBaseline_Update(sensor, detConf);
    }
    TSI_LTABaseline_Update(sensor, detConf);
#else
    /* Update baseline */
    TSI_NormalBaseline_Update(sensor, detConf);
#endif  /* if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U) */

    /* Calculate diffCount. */
    TSI_Sensor_UpdateDiffCount(sensor, detConf);
}

/* Private function implemenations ------------------------------------------*/
TSI_STATIC void TSI_Widget_ProcessDiffAndBaseline(TSI_LibHandleTypeDef *handle, TSI_WidgetTypeDef *widget)
{
    TSI_DetectConfTypeDef *widgetDetConf = &widget->detConf;
    uint16_t realSnsNum;

    TSI_UNUSED(handle)

    /* If it is self-cap parallel widget, we shall only init the first sensor. */
    if(TSI_WIDGET_IS_SELF_CAP(widget) && widget->meta->dedicatedScanGroup != NULL) {
        realSnsNum = 1U;
    }
    else {
        realSnsNum = widget->meta->sensorNum;
    }

    /* Update all sensors */
    TSI_FOREACH_OBJ(TSI_SensorTypeDef *, pSensor, widget->meta->sensors,
                    realSnsNum) {
        TSI_DetectConfTypeDef *detConf = pSensor->meta->detConf;
        if(detConf == NULL) { detConf = widgetDetConf; }
        TSI_ASSERT(detConf);
        TSI_Sensor_ProcessDiffAndBaseline(pSensor, detConf);
    }
    TSI_FOREACH_END()
}

TSI_STATIC void TSI_Widget_ProcessStatusAndBaseline(TSI_LibHandleTypeDef *handle, TSI_WidgetTypeDef *widget)
{
    TSI_DetectConfTypeDef *widgetDetConf = &widget->detConf;
    uint8_t sensorActive = 0U;
    uint16_t realSnsNum;

    TSI_UNUSED(handle)

    /* If it is self-cap parallel widget, we shall only init the first sensor. */
    if(TSI_WIDGET_IS_SELF_CAP(widget) && widget->meta->dedicatedScanGroup != NULL) {
        realSnsNum = 1U;
    }
    else {
        realSnsNum = widget->meta->sensorNum;
    }

    /* Update all sensors */
    TSI_FOREACH_OBJ(TSI_SensorTypeDef *, pSensor, widget->meta->sensors,
                    realSnsNum) {
        TSI_DetectConfTypeDef *detConf = pSensor->meta->detConf;
        if(detConf == NULL) { detConf = widgetDetConf; }
        TSI_ASSERT(detConf);
        TSI_Sensor_ProcessStatusAndBaseline(pSensor, detConf);
        if((pSensor->status & TSI_SENSOR_STATUS_ACTIVE) != 0U) {
            /* If any sensor is activated, mark widget status as active */
            sensorActive = 1U;
        }
    }
    TSI_FOREACH_END()

    /* Update widget general status */
    widget->status = sensorActive;
}

TSI_STATIC void TSI_Widget_ProcessPrivateData(TSI_LibHandleTypeDef *handle,
        TSI_WidgetTypeDef *widget)
{
    TSI_UNUSED(handle)

    /* Update widget-specified status */
    if(TSI_WIDGET_IS_SELF_CAP(widget)) {
        switch(widget->meta->type) {
            case TSI_WIDGET_SELF_CAP_BUTTON:
                TSI_Widget_UpdateSelfCapButton((TSI_SelfCapButtonTypeDef *)widget);
                break;

            case TSI_WIDGET_SELF_CAP_PROXIMITY:
                TSI_Widget_UpdateSelfCapProximity((TSI_SelfCapProximityTypeDef *)widget);
                break;

            case TSI_WIDGET_SELF_CAP_SLIDER:
                TSI_Widget_UpdateSelfCapSlider((TSI_SelfCapSliderTypeDef *)widget);
                break;

            case TSI_WIDGET_SELF_CAP_TOUCHPAD:
                TSI_Widget_UpdateSelfCapTouchPad((TSI_SelfCapTouchPadTypeDef *)widget);
                break;

            default:
                /* Cannot be here */
                TSI_ASSERT(0U);
                break;
        }
    }
    else if(TSI_WIDGET_IS_MUTUAL_CAP(widget)) {
        switch(widget->meta->type) {
            case TSI_WIDGET_MUTUAL_CAP_BUTTON:
                TSI_Widget_UpdateMutualCapButton((TSI_MutualCapButtonTypeDef *)widget);
                break;

            case TSI_WIDGET_MUTUAL_CAP_SLIDER:
                TSI_Widget_UpdateMutualCapSlider((TSI_MutualCapSliderTypeDef *)widget);
                break;

            default:
                /* Cannot be here */
                TSI_ASSERT(0U);
                break;
        }
    }
}

TSI_STATIC void TSI_Sensor_ProcessDiffAndBaseline(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    /*
        NOTE:
        This function MUST be used with TSI_Sensor_ProcessStatusAndBaseline() in pairs,
        or sensor-related values will be incorrected.
    */
    /* Pass rawCount through filter */
    TSI_Filter_Update(sensor);

    /* Update baseline with filtered rawCount */
    TSI_Baseline_Update(sensor, detConf);
}

TSI_STATIC void TSI_Sensor_ProcessStatusAndBaseline(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    /*
        NOTE:
        This function MUST be used with TSI_Sensor_ProcessDiffAndBaseline() in pairs,
        or sensor-related values will be incorrected.
    */
    TSI_WidgetTypeDef *widget = sensor->meta->parent;

    /* Update sensor active status */
    TSI_Sensor_UpdateStatus(sensor, detConf, TSI_SENSOR_STATUS_ACTIVE);

    if(widget->meta->type == TSI_WIDGET_SELF_CAP_PROXIMITY) {
        /* If it's proximity widget sensor, update sensor proxmity status */
        TSI_Sensor_UpdateStatus(sensor, detConf, TSI_SENSOR_STATUS_PROX);
    }

#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))
    /* Update baseline mode */
    TSI_Baseline_ModeJudge(sensor, detConf);
#endif  /* if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U) */
}

TSI_STATIC void TSI_Sensor_UpdateStatus(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf,
                                        uint8_t type)
{
    int32_t threshold;
    uint16_t hysteresis;
    uint8_t *pDebCount;
    uint8_t newStatus;
    TSI_WidgetTypeDef *widget = sensor->meta->parent;

    /* Get threshold */
    switch(type) {
        case TSI_SENSOR_STATUS_ACTIVE:
            threshold = detConf->activeTh;
            hysteresis = detConf->activeHys;
            pDebCount = &sensor->meta->debArray[0];
            break;

        case TSI_SENSOR_STATUS_PROX:
            threshold = ((TSI_SelfCapProximityTypeDef *)widget)->proxTh;
            hysteresis = ((TSI_SelfCapProximityTypeDef *)widget)->proxHys;
            pDebCount = &sensor->meta->debArray[1];
            break;

        default:
            /* Type not exists */
            TSI_ASSERT(0U);
            return;
    }

    /* Add hysteresis to threshold */
    threshold = ((sensor->status & type) == 0U) ?
                (threshold + hysteresis) :
                (threshold - hysteresis);

    /* Add debounce and update status */
    newStatus = (sensor->diffCount > threshold) ? type : 0U;
    if((newStatus != (sensor->status & type)) && ((*pDebCount) != 0U)) {
        if(--(*pDebCount) == 0U) {
            sensor->status &= ~type;
            sensor->status |= newStatus;
        }
    }
    else {
#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))
        if(sensor->bslnVar.bslnMode == TSI_BASELINE_MODE_NORMAL) {
            *pDebCount = ((sensor->status & type) == 0U) ?
                         detConf->onDebounce :
                         detConf->offDebounce;
        }
        else if(sensor->bslnVar.bslnMode == TSI_BASELINE_MODE_LTA) {
            *pDebCount = ((sensor->status & type) == 0U) ?
                         detConf->ltaOnDebounce :
                         detConf->offDebounce;
        }
        else {
            /* Cannot be here */
            TSI_ASSERT(0U);
            return;
        }
#else
        *pDebCount = ((sensor->status & type) == 0U) ?
                     detConf->onDebounce :
                     detConf->offDebounce;
#endif  /* if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U) */
    }
}

TSI_STATIC void TSI_NormalBaseline_Init(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    uint32_t freq;

    for(freq = 0U; freq < TSI_TOTAL_SCAN_NUM; freq++) {
        sensor->baseline[freq] = sensor->rawCount[freq];
        sensor->bslnVar.bslnNegStopCount[freq] = 0U;
        TSI_Filter_IIRInit(&sensor->bslnVar.bslnIIRBuff[freq], sensor->rawCount[freq]);
    }
}

TSI_STATIC void TSI_NormalBaseline_Update(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    TSI_WidgetTypeDef *widget = sensor->meta->parent;
    uint32_t freq;

    for(freq = 0U; freq < TSI_TOTAL_SCAN_NUM; freq++) {
        int32_t diffCount = (int32_t)sensor->rawCount[freq] - (int32_t)sensor->baseline[freq];
        if(diffCount >= 0) {
            sensor->bslnVar.bslnNegStopCount[freq] = 0U;
        }

        /* Reset baseline when negative stop timeout */
        if(diffCount < -((int32_t)widget->detConf.negNoiseTh)) {
            if(sensor->bslnVar.bslnNegStopCount[freq] >= widget->detConf.bslnNegStopTimeout) {
                TSI_NormalBaseline_Init(sensor, detConf);
            }
            else {
                sensor->bslnVar.bslnNegStopCount[freq]++;
            }
        }
        else {
#if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U)
            /*
            When sensor's auto-reset is disabled, baseline only
            updates when diffCount <= noiseTh
            */
            if((diffCount <= (int32_t)detConf->noiseTh) &&
                    diffCount >= -((int32_t)detConf->negNoiseTh)) {
#endif
                uint16_t tmp = sensor->rawCount[freq];
                TSI_Filter_IIRUpdate(&sensor->bslnVar.bslnIIRBuff[freq],
                                     detConf->bslnIIRCoeff, &tmp);
                sensor->baseline[freq] = tmp;
#if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U)
            }
#endif
        }
    }
}

#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))

TSI_STATIC void TSI_LTABaseline_Init(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    uint32_t freq;

    TSI_UNUSED(detConf)

    for(freq = 0U; freq < TSI_TOTAL_SCAN_NUM; freq++) {
        /* LTA baseline share normal baseline buffer */
        sensor->bslnVar.lta[freq] = sensor->rawCount[freq];
        TSI_Filter_IIRInit(&sensor->bslnVar.ltaIIRBuff[freq], sensor->rawCount[freq]);
    }
    sensor->bslnVar.bslnMode = TSI_BASELINE_MODE_NORMAL;
    sensor->bslnVar.ltaActiveCnt = 0U;
    sensor->bslnVar.ltaBslnStopCnt = 0U;
    sensor->bslnVar.ltaNegErrDebCnt = 0U;
}

TSI_STATIC void TSI_LTABaseline_Update(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    uint32_t freq;

    for(freq = 0U; freq < TSI_TOTAL_SCAN_NUM; freq++) {
        uint16_t tmp = sensor->rawCount[freq];
        TSI_Filter_IIRUpdate(&sensor->bslnVar.ltaIIRBuff[freq],
                             detConf->bslnIIRCoeff, &tmp);
        sensor->bslnVar.lta[freq] = tmp;
    }
}

TSI_STATIC void TSI_Baseline_ModeJudge(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    TSI_BaselineVarTypeDef *bslnVar = &sensor->bslnVar;
    uint32_t freq;

    if(bslnVar->bslnMode == TSI_BASELINE_MODE_NORMAL) {
        /* Sensor active timeout */
        if((sensor->status & TSI_SENSOR_STATUS_ACTIVE) != 0U) {
            bslnVar->ltaActiveCnt++;
        }
        else {
            bslnVar->ltaActiveCnt = 0U;
            /* Normal baseline positive stop timeout */
            if(sensor->diffCount > 0) {
                bslnVar->ltaBslnStopCnt++;
            }
            else {
                bslnVar->ltaBslnStopCnt = 0U;
            }
        }

        /* Update LTA baseline state */
        if((bslnVar->ltaActiveCnt > detConf->ltaActiveTimeout) ||
                (bslnVar->ltaBslnStopCnt > detConf->ltaNormBslnStopTimeout)) {
            /* Sensor active timeout / baseline stop timeout, enter LTA mode */
            bslnVar->bslnMode = TSI_BASELINE_MODE_LTA;
            bslnVar->ltaActiveCnt = 0U;
            bslnVar->ltaBslnStopCnt = 0U;
            memset(sensor->meta->debArray, detConf->onDebounce, sensor->meta->debArraySize);
        }
    }
    else if(bslnVar->bslnMode == TSI_BASELINE_MODE_LTA) {
        if((sensor->status & TSI_SENSOR_STATUS_ACTIVE) != 0U) {
            /* Sensor active: Raise baseline and switch back to normal baseline mode */
            for(freq = 0U; freq < TSI_TOTAL_SCAN_NUM; freq++) {
                sensor->baseline[freq] = sensor->bslnVar.lta[freq];
                sensor->bslnVar.bslnNegStopCount[freq] = 0U;
                TSI_Filter_IIRInit(&sensor->bslnVar.bslnIIRBuff[freq], sensor->bslnVar.lta[freq]);
            }
            bslnVar->bslnMode = TSI_BASELINE_MODE_NORMAL;
        }
        else {
            if(sensor->ltaDiffCount < -(int32_t)detConf->ltaNegErrTh) {
                if(++bslnVar->ltaNegErrDebCnt >= detConf->ltaNegErrorDebounce) {
                    /* LTA reset on negative error */
                    TSI_LTABaseline_Init(sensor, detConf);
                    memset(sensor->meta->debArray, detConf->onDebounce, sensor->meta->debArraySize);
                    bslnVar->ltaNegErrDebCnt = 0U;
                }
            }
            else {
                bslnVar->ltaNegErrDebCnt = 0U;
                if(sensor->diffCount < detConf->noiseTh) {
                    /* Normal baseline recover */
                    bslnVar->bslnMode = TSI_BASELINE_MODE_NORMAL;
                }
            }
        }
    }
    else {
        /* Cannot be here */
        TSI_ASSERT(0U);
        return;
    }
}

#endif  /* if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U) */

TSI_STATIC void TSI_Sensor_UpdateDiffCount(TSI_SensorTypeDef *sensor, TSI_DetectConfTypeDef *detConf)
{
    int32_t diffCount;
#if (TSI_SCAN_FREQ_NUM == 3U)
    int32_t diffCountMulti_0, diffCountMulti_2;
#endif  /* if (TSI_SCAN_FREQ_NUM == 3U) */

    /* NOTE: Will not update user-defined scan diffCount. Users should
       process it by themselves. */

    /*---------------------------------------------------#/
        Update normal baseline diffcount.
    #----------------------------------------------------*/
    diffCount = (int32_t)sensor->rawCount[0U] - (int32_t)sensor->baseline[0U];
#if (TSI_SCAN_FREQ_NUM == 3U)
    TSI_DEBUG("sensor(Tx%02d Rx%02d):", sensor->meta->txChannel, sensor->meta->rxChannel);
    TSI_DEBUG("- diff #0: %d", diffCount);
    diffCountMulti_0 = (int32_t)sensor->rawCount[1U] - (int32_t)sensor->baseline[1U];
    diffCountMulti_2 = (int32_t)sensor->rawCount[2U] - (int32_t)sensor->baseline[2U];
    TSI_DEBUG("- diff #1: %d", diffCountMulti_0);
    TSI_DEBUG("- diff #2: %d", diffCountMulti_2);

    /* Take median value as real diffCount */
    if(diffCountMulti_0 < diffCountMulti_2) {
        int32_t swap = diffCountMulti_0;
        diffCountMulti_0 = diffCountMulti_2;
        diffCountMulti_2 = swap;
    }
    if(diffCountMulti_0 > diffCount) {
        if(diffCount < diffCountMulti_2) {
            diffCount = diffCountMulti_2;
        }
    }
    else {
        diffCount = diffCountMulti_0;
    }

    TSI_DEBUG("- select: %d", diffCount);

#endif  /* if (TSI_SCAN_FREQ_NUM == 3U) */

    sensor->diffCount = 0U;
    if(diffCount > (int32_t)detConf->noiseTh) {
        sensor->diffCount = diffCount;
    }

#if((TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U))

    /*---------------------------------------------------#/
        Update LTA baseline diffcount.
    #----------------------------------------------------*/
    diffCount = (int32_t)sensor->rawCount[0U] - (int32_t)sensor->bslnVar.lta[0U];
#if (TSI_SCAN_FREQ_NUM == 3U)
    diffCountMulti_0 = (int32_t)sensor->rawCount[1U] - (int32_t)sensor->bslnVar.lta[1U];
    diffCountMulti_2 = (int32_t)sensor->rawCount[2U] - (int32_t)sensor->bslnVar.lta[2U];

    /* Take median value as real diffCount */
    if(diffCountMulti_0 < diffCountMulti_2) {
        int32_t swap = diffCountMulti_0;
        diffCountMulti_0 = diffCountMulti_2;
        diffCountMulti_2 = swap;
    }
    if(diffCountMulti_0 > diffCount) {
        if(diffCount < diffCountMulti_2) {
            diffCount = diffCountMulti_2;
        }
    }
    else {
        diffCount = diffCountMulti_0;
    }
#endif  /* if (TSI_SCAN_FREQ_NUM == 3U) */
    sensor->ltaDiffCount = diffCount;
#endif  /* if (TSI_SENSOR_BSLN_ALWAYS_UPDATE == 0U) && (TSI_SENSOR_BSLN_USE_LTA == 1U) */
}
